Unit dukeutil;
{
Utilities for interfacing Turbo Pascal V5.5 to Duke Nukem 3D v1.3

Revision History
Ver  Date      by   comments
2.0  22-Apr-98 JMH  Uses block file access, handles full size files
2.1  26-Apr-98 JMH  Added sort_walls to make sector blocks contiguous
2.2  10-Aug-98 JMH  removed pause before exit
2.3  30-Aug-98 JMH  improved error exit for memory overflow and scratchfile
}

Interface
Uses Dukedefs,Dos;

const absmaxsectors = 1024;
      absmaxwalls = absmaxsectors*8;
      absmaxsprites = absmaxsectors*4;

type sector_pointer_type = ^sectortype;
     sector_array_type = array[0..absmaxsectors-1] of sector_pointer_type;
     wall_pointer_type = ^walltype;
     wall_array_type = array[0..absmaxwalls-1] of wall_pointer_type;
     sprite_pointer_type = ^spritetype;
     sprite_array_type = array[0..absmaxsprites-1] of sprite_pointer_type;

Var maxsectors,maxwalls,maxsprites: integer;
    mapvsn: string[4];
    startposx,startposy,startposz: longint;
    startang,startsectnum: integer;
    numsectors,numwalls,numsprites: word;
    sector: sector_array_type;
    wall: wall_array_type;
    sprite: sprite_array_type;

Procedure NewSector(var s: integer; src: integer);
Procedure NewWall(var w: integer; src: integer);
Procedure NewSprite(var e: integer; src: integer);
Procedure Parse_filename (var s1: string; s: string);
Procedure LoadNewFile(var filename: string);
Procedure SaveNewFile(var filename: string);
Procedure Error(S: string);

Implementation

Procedure Overflow_Error(s: string; a,b: integer);
{ called to terminate program if there is not enough memory to create 's' }
begin
   Writeln('Error - insufficient memory for ',s,' storage');
   Writeln('Available = ',b,', required = ',a);
   Halt;
end;

Procedure NewSector (var s: integer; src: integer);
{ clones src sector and returns sector number }
begin
  if numsectors>=maxsectors then
     Overflow_Error('sector',numsectors+1,maxsectors);
  s:=numsectors;
  inc(numsectors);
  sector[s]^:=sector[src]^;
end;

Procedure NewWall(var w: integer; src: integer);
{ clones src wall and returns wall number }
begin
   if numwalls>=maxwalls then
      Overflow_Error('wall',numwalls+1,maxwalls);
   w:=numwalls;
   inc(numwalls);
   wall[w]^:=wall[src]^;
end;

Procedure NewSprite(var e:  integer; src: integer);
{ clones src sprite and returns sprite number }
begin
  if numsprites>=maxsprites then
     Overflow_Error('sprite',numsprites+1,maxsprites);
  e:= numsprites;
  inc(numsprites);
  sprite[e]^:=sprite[src]^;
end;


Procedure Parse_filename (var s1: string; s: string);
{ add s to s1 if it does not contain dot }
{ 20-Aug-98 improved path analysis }
var i: integer;
begin
   s1:=Fexpand(s1);
   i:=pos('.',s1);
   If i>0 then exit;
   s1:=s1+'.'+s;
end;

Procedure LoadNewFile(var filename: string);
{ loads current map from specified filename }
var i: integer;
    f: file;

begin
   Parse_filename (filename,'MAP');
   assign(f,filename);
   {$I-}
   reset(f,1);
   if (IOResult<>0) then Error('File not found: '+filename);
   {$I+}

   BlockRead(f,mapvsn,4);
   BlockRead(f,startposx,4);
   BlockRead(f,startposy,4);
   BlockRead(f,startposz,4);
   BlockRead(f,startang,2);
   BlockRead(f,startsectnum,2);

   BlockRead(f,numsectors,2);
   If numsectors>maxsectors then
      Overflow_Error('sector',numsectors,maxsectors);
   if numsectors>0 then
     for i:=0 to numsectors-1 do
       BlockRead(f,sector[i]^,sizeof(sectortype));

   BlockRead(f,numwalls,2);
   If numwalls>maxwalls then
      Overflow_Error('wall',numwalls,maxwalls);
   if numwalls>0 then
     for i:=0 to numwalls-1 do
       BlockRead(f,wall[i]^,sizeof(walltype));

   BlockRead(f,numsprites,2);
   If numsprites>maxsprites then
      Overflow_Error('sprite',numsprites,maxsprites);
   if numsprites>0 then
     for i:=0 to numsprites-1 do
       BlockRead(f,sprite[i]^,sizeof(spritetype));

   close(f);
end;

Procedure Squeeze_sprites;
{ removes all null sprites from sprite array }
var s,d : integer;
begin
   if numsprites=0 then exit;
   d:=0;
   for s:=0 to numsprites-1 do begin
      if sprite[s]^.sectnum>=0 then begin
         if d<s then sprite[d]^:=sprite[s]^;
         inc(d);
      end;
   end;
   numsprites:=d;
end;

Procedure Sort_walls;
{ generate list of walls in order and renumber all wall and sector pointers }
{ this uses a spare field in the wall definition as a scratch area      }
{ this is neededprior to saving the map file because the game engine    }
{ requires all walls in a sector to be contiguous }

var wall_count: integer;

Procedure Process_sector(first_wall,nwalls: integer);
{ subprocedure of sort_walls - creates list of new wall numbers }
var w,loop_start,i : integer;
begin
    loop_start:= first_wall;
    w:=loop_start;
    for i:=0 to nwalls-1 do begin
      if (w<0) or (w>=numwalls) then Error('wall list corrupt');

      wall[w]^.extra:=wall_count;
      inc(wall_count);

      with wall[w]^ do if point2=loop_start then begin
          inc(w);
          loop_start:=w;
      end else
          w:=point2;
   end;
end;

Procedure renumber ( var n: integer);
{ sub procedure of sort_walls }
begin
   if (n<0) or (n>=numwalls) then error('wall renumbering');
   n:=wall[n]^.extra;
end;

Procedure reorder_wall_pointers;
{ subprocedure of sort_walls - put all wall pointers into order }
{ use scratch file to avoid heap or stack overflow }

const scratch_fn='TEMP.$$$';

var i: integer;
    f: file of wall_pointer_type;
    exptr: wall_pointer_type;

begin
   assign(f,scratch_fn);
   {$I-}
   Rewrite(f);
   if IOResult<>0 then Error('Cannot create file: '+scratch_fn);
   {$I+}

   for i:=0 to numwalls-1 do write(f,wall[i]);
   reset(f);
   for i:=0 to numwalls-1 do begin
      read(f,exptr);
      wall[exptr^.extra]:=exptr;
   end;
   erase(f);
end;

var i: integer;

begin { sort_walls }
{ build indirection table for new sequence of walls }
   wall_count:=0;
   for i:=0 to numsectors-1 do
       with sector[i]^ do
          Process_sector(wallptr,wallnum);
   if wall_count<>numwalls then Error('wall list size error');

{ use indirection table to renumber wall pointers in sectors and walls }
   for i:=0 to numsectors-1 do renumber(sector[i]^.wallptr);
   for i:=0 to numwalls-1 do with wall[i]^ do begin
       if nextwall>0 then renumber(nextwall);
       renumber(point2);
   end;

{ use indirection table to reorder pointers to wall storage }
   Reorder_wall_pointers;
end; { sort_walls }


Procedure SaveNewfile(var filename: string);
{ Writes current map to specified file name }
{ empty sprites are squeezed }
Var i: integer;
    f: file;

begin
   Sort_walls;
   Squeeze_sprites;

   Parse_filename(filename,'MAP');
   Assign(f,filename);
   {$I-}
   Rewrite(f,1);
   if IOResult<>0 then Error('Cannot create file: '+filename);
   {$I+}

   BlockWrite(f,mapvsn,4);
   BlockWrite(f,startposx,4);
   BlockWrite(f,startposy,4);
   BlockWrite(f,startposz,4);
   BlockWrite(f,startang,2);
   BlockWrite(f,startsectnum,2);

   BlockWrite(f,numsectors,2);
   if numsectors>0 then
     for i:=0 to numsectors-1 do
       BlockWrite(f,sector[i]^,sizeof(sectortype));

   BlockWrite(f,numwalls,2);
   if numwalls>0 then
     for i:=0 to numwalls-1 do
       BlockWrite(f,wall[i]^,sizeof(walltype));

   BlockWrite(f,numsprites,2);
   if numsprites>0 then
     for i:=0 to numsprites-1 do
       BlockWrite(f,sprite[i]^,sizeof(spritetype));

   close(f);
end;

Procedure Error(S: string);
{ universal error exit routine }
begin
   Writeln('Error - ',s);
   Halt;
end;

{ Initialisation at program startup }

var i: integer;
begin
{ allocate all but 32K of the heap to sector, wall and sprite records }
  maxsectors:=(Memavail-32000) div (sizeof(sectortype)
             +8*sizeof(walltype)+4*sizeof(spritetype));
  if maxsectors>absmaxsectors then maxsectors:=absmaxsectors;

{ now allocate heap space and fix pointers for sector type .. }
  for i:=0 to maxsectors-1 do new(sector[i]);
  numsectors:=0;

{ .. ditto for wall type .. }
  maxwalls:=maxsectors*8;
  for i:=0 to maxwalls-1 do new(wall[i]);
  numwalls:=0;

{ .. and finally for sprites type }
  maxsprites:=maxsectors*4;
  for i:=0 to maxsprites-1 do new(sprite[i]);
  numsprites:=0;
end.
